#!/usr/bin/env python
# SPDX-License-Identifier: MIT
from os.path import basename, expanduser
from typing import List, Optional

from absl import app, flags
from progress.bar import Bar

from .config import ToolConfig
from .device import Device, list_devices
from .types import *

FLAGS = flags.FLAGS
flags.DEFINE_string('config_file', '~/.busytag.toml', 'Config file path')
flags.DEFINE_string('device', None, 'Busy Tag\'s serial port.')
flags.DEFINE_integer('baudrate', 115200, 'Connection baudrate.')

class ProgressBar(Bar, ProgressListener):
    suffix = '%(position_bytes)s / %(total_bytes)s - %(eta)ds'

    @property
    def position_bytes(self) -> str:
        return format_size(self.index)

    @property
    def total_bytes(self) -> str:
        return format_size(self.max)

    def set_max(self, max: int) -> None:
        self.max = max
        self.start()


def format_size(size: int) -> str:
    if size < 1_000:
        return f'{size} B'
    if size < 500_000:
        return f'{size / 1_000:.2f} kB'
    return f'{size / 1_000_000:.2f} MB'


def main(argv: List[str]) -> Optional[int]:
    config = ToolConfig(FLAGS.config_file)
    if FLAGS.device is not None:
        config.device = FLAGS.device
    config.write_to_file()
    bt: Optional[Device] = None

    # Remove argv[0]
    exec_name = argv.pop(0)
    command = 'help'
    if len(argv) > 0:
        command = argv.pop(0)

        # Don't bother connecting for commands that don't need a device connection.
        if command not in ('list_devices', 'help'):
            if config.device is None:
                print(
                    'A device must be specified using the `--device` flag or be set in the config file!')
                print(
                    f'You can run `{exec_name} list_devices` to find the available devices.')
                return 1
            bt = Device(config.device, baudrate=FLAGS.baudrate)

    match command:
        case 'info':
            print(f'Device name:      {bt.name}')
            print(f'Device ID:        {bt.device_id}')
            print(f'Firmware version: {bt.firmware_version}')
            print(f'Serial port:      {config.device}')
            print(f'Storage capacity: {format_size(bt.capacity)}')
            print(f'Free storage:     {format_size(bt.get_free_storage())}')

        case 'list_devices':
            devices = list_devices(FLAGS.baudrate)
            if len(devices) == 0:
                print('No devices found')
            else:
                print('Available devices:')
                for device in devices:
                    print(f'  {device}')

        case 'list_pictures':
            print('Pictures in device:')
            for picture in bt.list_pictures():
                print(f'  {picture.name} ({format_size(picture.size)})')
            print(f'Available space: {format_size(bt.get_free_storage())}')

        case 'list_files':
            print('Files in device: ')
            for file in bt.list_files():
                print(
                    f'  {file.name} ({file.type.value} - {format_size(file.size)})')
            print(f'Available space: {format_size(bt.get_free_storage())}')

        case 'set_picture':
            assert len(argv) >= 1
            bt.set_active_picture(argv.pop(0))

        case 'get_picture':
            print(f'Current active picture: {bt.get_active_picture()}')

        case 'put':
            assert len(argv) >= 1
            filename = expanduser(argv.pop(0))
            with open(filename, 'rb') as fp:
                bt.upload_file(basename(filename), fp.read(),
                               ProgressBar(f'Uploading {basename(filename)}'))

        case 'get':
            assert len(argv) >= 1
            filename = argv.pop(0)
            data = bt.read_file(filename, ProgressBar(f'Downloading {filename}'))
            with open(filename, 'wb') as fp:
                fp.write(data)

        case 'rm':
            assert len(argv) >= 1
            filename = argv.pop(0)
            bt.delete_file(filename)

        case 'set_led_solid_color':
            assert len(argv) >= 1
            led_config = LedConfig(LedPin.ALL, argv.pop(0).upper())
            bt.set_led_solid_color(led_config)

        case 'apply_led_preset':
            assert len(argv) >= 1
            preset_name = argv.pop(0)

            # Clear all LEDs first
            bt.set_led_solid_color(LedConfig(LedPin.ALL, '000000'))
            for e in config.led_presets.get(preset_name):
                bt.set_led_solid_color(e)

        case 'get_brightness':
            print(f'Brightness: {bt.get_display_brightness()}')

        case 'set_brightness':
            assert len(argv) >= 1
            brightness = int(argv.pop(0))
            assert 0 < brightness <= 100
            bt.set_display_brightness(brightness)

        case 'help':
            print(f'\n\tUSAGE: {exec_name} [flags] <command> [<args>]\n')
            print('Available commands:')
            print('  help: Prints this message')
            print('  list_devices: Lists available devices')
            print('  info: Displays device information')
            print('  list_pictures: Lists pictures in device')
            print('  list_files: Lists files in device')
            print('  get_picture: Gets the filename of the picture being shown')
            print(
                '  set_picture <filename>: Sets the picture shown in the device')
            print('  put <filename>: Uploads <filename>')
            print(
                '  get <filename>: Copies <filename> from the device to the working directory')
            print('  rm <filename>: Deletes <filename>')
            print(
                '  set_led_solid_color <6 hex RGB colour>: Sets the LEDs colour')
            print(
                '  apply_led_preset <preset name>: Sets the LEDs colour according to a preset')
            print('  get_brightness: Gets current display brightness')
            print(
                '  set_brightness <brightness>: Sets current display brightness (int between 1 and 100, inclusive')
            print(f'\nFor flag documentation, run {exec_name} --help')

        case _:
            print(
                f'Unknown command `{command}`. Please use the `help` to list available commands')
            return 1

    return 0


def run_main():
    app.run(main)


if __name__ == '__main__':
    run_main()



# SPDX-License-Identifier: MIT

from typing import Sequence, Optional, List, Iterator

import serial
from absl import logging
from serial.tools.list_ports import comports

from .types import *

__all__ = ['Device', 'list_devices']


def generate_chunks(n: int) -> Iterator[int]:
    chunk_size = 1_000
    bytes_chunked = 0
    while bytes_chunked + chunk_size < n:
        yield chunk_size
        bytes_chunked += chunk_size

    if n % chunk_size != 0:
        yield n % chunk_size


def list_devices(baudrate: int) -> List[str]:
    """Lists all Busy Tag devices connected to the computer.

    This function tries to connect to all serial devices that report being a 'BUSY TAG' product, and returns a list of
    those that responded to the AT+GDN command according to the API documentation.

    :param baudrate: Baudrate to use when connecting to the devices.
    :return: List of port names for the detected devices.
    """
    devices = []
    ports = comports()
    for port in ports:
        if port.product != 'BUSY TAG':
            continue

        try:
            logging.debug(f'Trying to connect to {port.device}')
            with serial.Serial(port.device, baudrate=baudrate,
                               timeout=1.0) as conn:
                logging.debug(f'Connected to {port.device}')
                conn.write(b'AT+GDN\r\n')
                while True:
                    response = conn.readline()
                    logging.debug(f'Read from device: {response}')
                    if response.startswith(b'+evn'):
                        continue
                    if response.startswith(b'+DN:busytag-'):
                        devices.append(port.device)
                    break
        except Exception:
            pass

    return devices


class Device(object):
    """Class to interact with Busy-Tag devices through a serial connection.

    The protocol to communicate with the Busy-Tag device is documented at
    https://luxafor.helpscoutdocs.com/article/47-busy-tag-usb-cdc-command-reference-guide.
    """

    def __init__(self, port_path: Optional[str] = None,
                 connection: Optional[serial.Serial] = None,
                 baudrate: int = 115200):
        assert not (port_path is None and connection is None)
        if port_path is not None:
            self._port = port_path
            logging.info(f'Connecting to serial port {port_path}')
            self.conn = serial.Serial(port_path, baudrate)
        else:
            self._port = None
            self.conn = connection

        self.__capacity = int(self.__get_readonly_attribute('TSS'))
        self.__device_id = self.__get_readonly_attribute('ID')
        self.__firmware_version = self.__get_readonly_attribute('FV')
        self.__hostname = self.__get_readonly_attribute('LHA').removeprefix(
            'http://')
        self.__manufacturer = self.__get_readonly_attribute('MN')
        self.__name = self.__get_readonly_attribute('DN')

    def list_pictures(self) -> Sequence[FileEntry]:
        """Lists pictures that can be displayed on the screen."""
        self.__send_command('AT+GPL')
        result = []
        while True:
            l = self.__readline()
            # Unlikely, but event messages might arrive while we're listing
            # files. Silently consume them.
            if l.startswith(b'+evn'):
                continue

            if l.startswith(b'OK'):
                break

            filename, size = l.decode().removeprefix('+PL:').split(',')
            result.append(FileEntry(filename, int(size)))

        return result

    def list_files(self) -> Sequence[FileEntry]:
        """Lists all files stored on the device."""
        self.__send_command('AT+GFL')
        result = []
        while True:
            l = self.__readline()
            if l.startswith(b'+evn'):
                continue
            if l.startswith(b'OK'):
                break
            filename, entry_type, size = l.decode().removeprefix('+FL:').split(
                ',')
            result.append(
                FileEntry(filename, int(size), FileEntryType(entry_type)))

        return result

    def read_file(self, filename: str, progress_listener: Optional[
        ProgressListener] = None) -> bytes:
        """Reads a file from the device.

        :param filename: The filename of the file to read
        :param progress_listener:
        :return: a `bytes` object with the file contents
        """
        logging.info(f'Reading file {filename}')

        self.__send_command('AT+GF=%s' % (filename,))
        # First part of response: +GF:<filename>,<size in bytes>\r\n
        response = self.__read_response('+GF:')
        if b',' not in response:
            raise IOError(
                'Malformed response to command AT+GF=%s' % (response,))

        read_size = int(response.split(b',')[1]) + 8
        logging.debug(f'Reading {read_size} bytes from device')
        if progress_listener is not None:
            progress_listener.set_max(read_size)
        response = b''
        for chunk_size in generate_chunks(read_size):
            response += self.conn.read(chunk_size)
            if progress_listener is not None:
                progress_listener.goto(len(response))

        if progress_listener is not None:
            progress_listener.finish()

        assert response[-6:] == b'\r\nOK\r\n'
        return response[2:-6]

    def upload_file(self, filename: str, data: bytes,
                    progress_listener: Optional[
                        ProgressListener] = None):
        """Uploads a file to the device.

        :param filename: The filename of the file to upload
        :param data: The contents of the file to upload
        :param progress_listener:
        """
        logging.info(f'Uploading file {filename} ({len(data)} bytes)')

        self.__send_command('AT+UF=%s,%d' % (filename, len(data)))
        self.__readline()
        logging.debug('Writing %d bytes to device', len(data))
        bytes_written = 0

        if progress_listener is not None:
            progress_listener.set_max(len(data))

        for chunk_size in generate_chunks(len(data)):
            self.conn.write(data[bytes_written:bytes_written + chunk_size])
            bytes_written += chunk_size
            if progress_listener is not None:
                progress_listener.goto(bytes_written)

        logging.debug('Waiting for device to finish')
        terminator = self.conn.read(6)

        if progress_listener is not None:
            progress_listener.finish()

        assert terminator == b'\r\nOK\r\n'

    def delete_file(self, filename: str):
        """Deletes a file from the device.'

        :param filename: The filename of the file to delete
        """
        logging.info(f'Deleting file {filename}')
        self.__send_command('AT+DF=%s' % (filename,))
        self.__read_response('+DF:')
        self.__read_response('OK')

    def set_active_picture(self, filename: str):
        """Set the picture that will be shown on the display."""
        logging.info(f'Setting active picture {filename}')
        self.__set_attribute('SP', filename)

    def get_active_picture(self) -> str:
        """Gets the file name of the picture being displayed."""
        return self.__get_attribute('SP')

    def get_free_storage(self) -> int:
        return int(self.__get_readonly_attribute('FSS'))

    def get_display_brightness(self) -> int:
        return int(self.__get_attribute('DB'))

    def set_display_brightness(self, brightness: int):
        if brightness < 1 or brightness > 100:
            raise ValueError('Brightness must be between 1 and 100')
        self.__set_attribute('DB', f'{brightness}')

    def get_led_solid_color(self) -> LedConfig:
        pins, rgb = self.__get_attribute('SC').split(',')
        return LedConfig(LedPin(int(pins)), rgb)

    def set_led_solid_color(self, config: LedConfig):
        self.__set_attribute('SC', f'{int(config.pins)},{config.color}')

    def get_led_pattern(self) -> Sequence[LedPatternEntry]:
        result = []
        self.__send_command('AT+CP?')
        while True:
            entry = self.__readline().strip()
            if entry.startswith(b'ERROR'):
                logging.error('Received error response: %s', entry)
                raise self.build_exception(entry)
            if entry.startswith(b'+CP'):
                pins, rgb, speed, delay = entry.removeprefix(
                    b'+CP:').decode().split(',')
                result.append(
                    LedPatternEntry(LedPin(int(pins)), rgb, int(speed),
                                    int(delay)))
            elif entry.startswith(b'OK'):
                break

        return result

    def set_led_pattern(self, pattern: Sequence[LedPatternEntry]):
        assert 0 < len(pattern) <= 40
        self.__send_command(f'AT+CP={len(pattern)}')
        self.__read_response('>')
        for entry in pattern:
            self.__send_command(f'+CP:{entry}')
        self.__read_response('OK')

    def get_wifi_config(self) -> WifiConfig:
        response = self.__get_attribute('WC')
        if ',' not in response:
            raise IOError(f'Malformed response to command AT+WC: {response}')
        ssid, password = response.split(',', 1)
        return WifiConfig(ssid, password)

    def set_wifi_config(self, wifi_config: WifiConfig):
        logging.info(f'Setting Wifi SSID to {wifi_config.ssid}')
        self.__set_attribute('WC', f'{wifi_config.ssid},{wifi_config.password}')

    def reset_wifi_config(self):
        logging.info('Resetting wifi configuration')
        self.__send_command('AT+FRWCF')
        self.__read_response('OK')

    @property
    def capacity(self) -> int:
        return self.__capacity

    @property
    def device_id(self) -> str:
        return self.__device_id

    @property
    def hostname(self) -> str:
        return self.__hostname

    @property
    def firmware_version(self) -> str:
        return self.__firmware_version

    @property
    def manufacturer(self) -> str:
        return self.__manufacturer

    @property
    def name(self) -> str:
        return self.__name

    def __get_readonly_attribute(self, attribute: str) -> str:
        """Retrieves the value of a read-only attribute (e.g. device id)."""
        response_prefix = f'+{attribute}:'
        self.__send_command(f'AT+G{attribute}')
        return self.__read_response(response_prefix).decode().removeprefix(
            response_prefix)

    def __get_attribute(self, attribute: str) -> str:
        """Retrieves the value of a user-modifiable attribute (e.g. active picture)."""
        response_prefix = f'+{attribute}:'
        self.__send_command(f'AT+{attribute}?')
        return self.__read_response(response_prefix).decode().removeprefix(
            response_prefix)

    def __set_attribute(self, attribute: str, value: str):
        """Sets the value of a user-modifiable attribute (e.g. active picture)."""
        self.__send_command(f'AT+{attribute}={value}')
        self.__read_response('OK')

    def __send_command(self, command: str):
        encoded_command = command.encode() + b'\r\n'
        logging.debug('Sending command: %s', encoded_command)
        self.conn.write(encoded_command)

    def __read_response(self, prefix: str) -> bytes:
        logging.debug(f'Waiting for prefix: {prefix}')
        encoded_prefix = prefix.encode()
        while True:
            response = self.__readline()
            if response.startswith(encoded_prefix):
                return response

    def __readline(self) -> bytes:
        result = self.conn.readline()
        logging.debug('Read from device: %s', result)
        if result.startswith(b'ERROR'):
            logging.error('Received error response: %s', result)
            raise self.build_exception(result)
        return result.strip()

    @staticmethod
    def build_exception(error_response: bytes) -> Exception:
        """Converts an error response from Busy Tag to an Exception"""
        if not error_response.startswith(b'ERROR:'):
            return Exception(
                'Unexpected error response %s' % (error_response.decode(),))

        parts = error_response.decode().strip().split(':')
        if len(parts) != 2:
            return Exception(
                'Unexpected error response %s' % (error_response.decode(),))

        match parts[1]:
            case '-1':
                return Exception(
                    'Unexpected error response %s' % (error_response.decode(),))
            case '0':
                return Exception('Unknown error')
            case '1':
                return ValueError('Invalid command')
            case '2':
                return ValueError('Invalid argument')
            case '3':
                return FileNotFoundError('File not found')
            case '4':
                return ValueError('Invalid size')
            case _:
                return Exception(
                    'Unexpected error response %s' % (error_response.decode(),))



# SPDX-License-Identifier: MIT

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, IntFlag, verify, NAMED_FLAGS

rgb_re = re.compile(r'^[0-9A-F]{6}$')

__all__ = ['LedPin', 'FileEntry', 'FileEntryType', 'WifiConfig', 'LedConfig',
           'LedPatternEntry', 'ProgressListener']


@verify(NAMED_FLAGS)
class LedPin(IntFlag):
    PIN_0 = 1
    PIN_1 = 2
    PIN_2 = 4
    PIN_3 = 8
    PIN_4 = 16
    PIN_5 = 32
    PIN_6 = 64
    ALL = 127


class FileEntryType(Enum):
    FILE = 'file'
    DIRECTORY = 'dir'


@dataclass(frozen=True)
class FileEntry:
    """File stored in the Busy Tag device."""
    name: str
    size: int
    type: FileEntryType = FileEntryType.FILE


@dataclass(frozen=True)
class WifiConfig:
    """Wifi configuration."""
    ssid: str
    password: str


@dataclass(frozen=True)
class LedConfig:
    pins: LedPin
    color: str

    def __post_init__(self):
        assert rgb_re.match(self.color)


@dataclass(frozen=True)
class LedPatternEntry:
    pins: LedPin
    color: str
    speed: int
    delay: int

    def __post_init__(self):
        assert rgb_re.match(self.color)
        assert self.speed >= 0
        assert self.delay >= 0

    def __str__(self) -> str:
        return f'{int(self.pins)},{self.color},{self.speed},{self.delay}'


class ProgressListener(ABC):
    """Interface passed to operations that (might) take a long time,
    which receives updates on the operation progress."""

    @abstractmethod
    def set_max(self, max: int) -> None:
        """Sets the size of the operation (e.g. number of bytes in a file operation).

        :param max: The max possible progress value."""
        pass

    @abstractmethod
    def goto(self, position: int):
        """Updates how far along the operation is.

        :param position: The current position of the operation."""
        pass

    @abstractmethod
    def finish(self):
        """Called when the operation has finished."""
        pass



# SPDX-License-Identifier: MIT

import os.path
from typing import Optional, Sequence

import tomlkit

from .types import *


class ToolConfig(object):
    def __init__(self, path: Optional[str] = None):
        self.device = None
        self.path = path
        self.led_presets = {}
        if path is not None:
            self.path = os.path.expanduser(path)
            if os.path.exists(self.path):
                self.__load_from_file()

    def write_to_file(self):
        assert self.path is not None
        conf = {}

        if self.device is not None:
            conf['device'] = self.device

        conf['led_presets'] = {}
        for name, settings in self.led_presets.items():
            preset = []
            for setting in settings:
                preset.append({'pins': int(setting.pins), 'color': setting.color})
            conf['led_presets'][name] = preset

        with open(self.path, "w") as fp:
            tomlkit.dump(conf, fp)

    def get_led_preset(self, name: str) -> Sequence[LedConfig]:
        if name not in self.led_presets:
            raise ValueError(f'LED preset {name} not found')
        return self.led_presets[name][::]

    def __load_from_file(self):
        with open(self.path, 'rb') as fp:
            conf = tomlkit.load(fp)
            if 'device' in conf:
                self.device = conf['device']

            if 'led_presets' in conf:
                for key, entry in conf['led_presets'].items():
                    self.led_presets[key] = []
                    for pattern in entry:
                        self.led_presets[key].append(LedConfig(LedPin(int(pattern['pins'])), pattern['color']))



import os

import pytest
import tomlkit
from busytag.config import ToolConfig
from busytag.types import LedConfig, LedPin


@pytest.fixture
def testdata_path():
    def _testdata_path(name):
        return os.path.join(os.path.dirname(__file__), 'testdata',
                            f"{name}.toml")

    return _testdata_path


def test_loads_config_with_device(testdata_path):
    config = ToolConfig(testdata_path('config-with-device'))
    assert config.device == '/dev/tty.usbmodem'


def test_loads_config_with_led_preset(testdata_path):
    config = ToolConfig(testdata_path('config-with-led-patterns'))
    expected_rb_pattern = {
        LedConfig(
            pins=LedPin.PIN_0 | LedPin.PIN_2 | LedPin.PIN_4 | LedPin.PIN_6,
            color='FF0000'),
        LedConfig(
            pins=LedPin.PIN_1 | LedPin.PIN_3 | LedPin.PIN_5,
            color='0000FF'
        )
    }

    diff = set(config.get_led_preset('rb')) ^ expected_rb_pattern
    assert not diff



def test_writes_config(tmp_path):
    p = tmp_path / 'config_with_device.toml'
    config = ToolConfig(str(p))

    assert config.device is None
    config.device = '/dev/tty.somedevice'
    config.write_to_file()

    c = tomlkit.loads(p.read_text())
    assert c['device'] == '/dev/tty.somedevice'
